/*****************************************************************************\

Copyright (c) Intel Corporation (2009-2017).

    INTEL MAKES NO WARRANTY OF ANY KIND REGARDING THE CODE.  THIS CODE IS
    LICENSED ON AN "AS IS" BASIS AND INTEL WILL NOT PROVIDE ANY SUPPORT,
    ASSISTANCE, INSTALLATION, TRAINING OR OTHER SERVICES.  INTEL DOES NOT
    PROVIDE ANY UPDATES, ENHANCEMENTS OR EXTENSIONS.  INTEL SPECIFICALLY
    DISCLAIMS ANY WARRANTY OF MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR ANY
    PARTICULAR PURPOSE, OR ANY OTHER WARRANTY.  Intel disclaims all liability,
    including liability for infringement of any proprietary rights, relating to
    use of the code. No license, express or implied, by estoppel or otherwise,
    to any intellectual property rights is granted herein.

  \file resource.h

\*****************************************************************************/

#ifndef __RESOURCE__
#define __RESOURCE__

#define OPENCL_C_H                     "OPENCL_C_H"
#define OPENCL_C_BASE_H                "OPENCL_C_BASE_H"
#define OPENCL_C_12_SPIR_PCM           "OPENCL_C_12_SPIR_PCM"
#define OPENCL_C_20_SPIR_PCM           "OPENCL_C_20_SPIR_PCM"
#define OPENCL_C_30_SPIR_PCM           "OPENCL_C_30_SPIR_PCM"
#define OPENCL_C_12_SPIR64_PCM         "OPENCL_C_12_SPIR64_PCM"
#define OPENCL_C_20_SPIR64_PCM         "OPENCL_C_20_SPIR64_PCM"
#define OPENCL_C_30_SPIR64_PCM         "OPENCL_C_30_SPIR64_PCM"
#define OPENCL_C_12_SPIR_FP64_PCM      "OPENCL_C_12_SPIR_FP64_PCM"
#define OPENCL_C_20_SPIR_FP64_PCM      "OPENCL_C_20_SPIR_FP64_PCM"
#define OPENCL_C_30_SPIR_FP64_PCM      "OPENCL_C_30_SPIR_FP64_PCM"
#define OPENCL_C_12_SPIR64_FP64_PCM    "OPENCL_C_12_SPIR64_FP64_PCM"
#define OPENCL_C_20_SPIR64_FP64_PCM    "OPENCL_C_20_SPIR64_FP64_PCM"
#define OPENCL_C_30_SPIR64_FP64_PCM    "OPENCL_C_30_SPIR64_FP64_PCM"
#define OPENCL_C_MODULE_MAP            "OPENCL_C_MODULE_MAP"

#endif /* __RESOURCE__ */
