// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

1 TEXTINCLUDE
BEGIN
    "resource.h\0"
END

//
// Module with OpenCL C declarations with corresponding headers
//

OPENCL_C_H                     PCM   "opencl-c.h"
OPENCL_C_BASE_H                PCM   "opencl-c-base.h"
OPENCL_C_12_SPIR_PCM           PCM   "opencl-c-12-spir.pcm"
OPENCL_C_20_SPIR_PCM           PCM   "opencl-c-20-spir.pcm"
OPENCL_C_30_SPIR_PCM           PCM   "opencl-c-30-spir.pcm"
OPENCL_C_12_SPIR64_PCM         PCM   "opencl-c-12-spir64.pcm"
OPENCL_C_20_SPIR64_PCM         PCM   "opencl-c-20-spir64.pcm"
OPENCL_C_30_SPIR64_PCM         PCM   "opencl-c-30-spir64.pcm"
OPENCL_C_12_SPIR_FP64_PCM      PCM   "opencl-c-12-spir-fp64.pcm"
OPENCL_C_20_SPIR_FP64_PCM      PCM   "opencl-c-20-spir-fp64.pcm"
OPENCL_C_30_SPIR_FP64_PCM      PCM   "opencl-c-30-spir-fp64.pcm"
OPENCL_C_12_SPIR64_FP64_PCM    PCM   "opencl-c-12-spir64-fp64.pcm"
OPENCL_C_20_SPIR64_FP64_PCM    PCM   "opencl-c-20-spir64-fp64.pcm"
OPENCL_C_30_SPIR64_FP64_PCM    PCM   "opencl-c-30-spir64-fp64.pcm"
OPENCL_C_MODULE_MAP            PCM   "module.modulemap"
